@@ -19,7 +19,7 @@ from __future__ import print_function
 
 __author__ = 'A. Jesse Jiryu Davis'
 __email__ = 'jesse@mongodb.com'
-__version__ = '1.7.0'
+__version__ = '1.8.0'
 
 import atexit
 import contextlib
@@ -27,6 +27,7 @@ import datetime
 import errno
 import functools
 import inspect
+import operator
 import os
 import random
 import select
@@ -39,6 +40,7 @@ import time
 import weakref
 import sys
 from codecs import utf_8_decode as _utf_8_decode
+from collections import OrderedDict
 
 try:
     from queue import Queue, Empty
@@ -50,11 +52,6 @@ try:
 except:
     from collections import Mapping
 
-try:
-    from collections import OrderedDict
-except:
-    from ordereddict import OrderedDict  # Python 2.6, "pip install ordereddict"
-
 try:
     from io import StringIO
 except ImportError:
@@ -203,9 +200,7 @@ class Future(object):
         self._event = threading.Event()
 
     def result(self, timeout=None):
-        self._event.wait(timeout)
-        # wait() always returns None in Python 2.6.
-        if not self._event.is_set():
+        if not self._event.wait(timeout):
             raise AssertionError('timed out waiting for Future')
         return self._result
 
@@ -275,7 +270,10 @@ REPLY_FLAGS = OrderedDict([
 
 OP_MSG_FLAGS = OrderedDict([
     ('checksumPresent', 1),
-    ('moreToCome', 2)])
+    ('moreToCome', 2),
+    ('exhaustAllowed', 16)])
+
+_ALL_OP_MSG_FLAGS = functools.reduce(operator.or_, OP_MSG_FLAGS.values())
 
 _UNPACK_BYTE = struct.Struct("<b").unpack
 _UNPACK_INT = struct.Struct("<i").unpack
@@ -628,42 +626,65 @@ class OpMsg(CommandBase):
         Takes the client message as bytes, the client and server socket objects,
         and the client request id.
         """
+        payload_document = OrderedDict()
         flags, = _UNPACK_UINT(msg[:4])
         pos = 4
-        first_payload_type, = _UNPACK_BYTE(msg[pos:pos + 1])
-        pos += 1
-        first_payload_size, = _UNPACK_INT(msg[pos:pos + 4])
-        if flags != 0 and flags != 2:
-            raise ValueError('OP_MSG flag must be 0 or 2 not %r' % (flags,))
-        if first_payload_type != 0:
-            raise ValueError('First OP_MSG payload type must be 0 not %r' % (
-                first_payload_type,))
+        if flags & ~_ALL_OP_MSG_FLAGS:
+            raise ValueError(
+                'OP_MSG flags has reserved bits set.'
+                ' Allowed flags: 0x%x. Provided flags: 0x%x' % (
+                    _ALL_OP_MSG_FLAGS, flags))
 
-        # Parse the initial document and add the optional payload type 1.
-        payload_document = bson.decode_all(msg[pos:pos + first_payload_size],
-                                           CODEC_OPTIONS)[0]
-        pos += first_payload_size
-        if len(msg) != pos:
+        checksum_present = flags & OP_MSG_FLAGS['checksumPresent']
+        checksum = None
+        if checksum_present:
+            msg_len_without_checksum = len(msg) - 4
+        else:
+            msg_len_without_checksum = len(msg)
+
+        while pos < msg_len_without_checksum:
             payload_type, = _UNPACK_BYTE(msg[pos:pos + 1])
             pos += 1
-            if payload_type != 1:
-                raise ValueError('Second OP_MSG payload type must be 1 not %r'
-                                 % (payload_type,))
-            section_size, = _UNPACK_INT(msg[pos:pos + 4])
-            if len(msg) != pos + section_size:
-                raise ValueError('More than two OP_MSG sections unsupported')
-            pos += 4
-            identifier, pos = _get_c_string(msg, pos)
-            documents = bson.decode_all(msg[pos:], CODEC_OPTIONS)
-            payload_document[identifier] = documents
+            payload_size, = _UNPACK_INT(msg[pos:pos + 4])
+            if payload_type == 0:
+                doc = bson.decode_all(msg[pos:pos + payload_size],
+                                      CODEC_OPTIONS)[0]
+                payload_document.update(doc)
+                pos += payload_size
+            elif payload_type == 1:
+                section_size, = _UNPACK_INT(msg[pos:pos + 4])
+                pos += 4
+                identifier, pos = _get_c_string(msg, pos)
+                # Section starts w/ 4-byte size prefix, identifier ends w/ nil.
+                documents_len = section_size - len(identifier) - 1 - 4
+                documents = bson.decode_all(msg[pos:pos + documents_len],
+                                            CODEC_OPTIONS)
+                payload_document[identifier] = documents
+                pos += documents_len
+
+        remaining = len(msg) - pos
+        if checksum_present:
+            if remaining != 4:
+                raise ValueError(
+                    'OP_MSG has checksumPresent flag set, expected 4 bytes'
+                    ' remaining but have %d bytes remaining' % (remaining,))
+
+            checksum = _UNPACK_UINT(msg[pos:pos+4])[0]
+        else:
+            if remaining != 0:
+                raise ValueError(
+                    'OP_MSG has no checksumPresent flag, expected 0 bytes'
+                    ' remaining but have %d bytes remaining' % (remaining,))
 
         database = payload_document['$db']
         return OpMsg(payload_document, namespace=database, flags=flags,
-                     _client=client, request_id=request_id,
+                     _client=client, request_id=request_id, checksum=checksum,
                      _server=server)
 
     def __init__(self, *args, **kwargs):
+        checksum = kwargs.pop('checksum', None)
         super(OpMsg, self).__init__(*args, **kwargs)
+        self._checksum = checksum
         if len(self._docs) > 1:
             raise_args_err('OpMsg too many documents', ValueError)
 
@@ -673,6 +694,11 @@ class OpMsg(CommandBase):
         read_preference = self.doc.get('$readPreference')
         return read_preference and read_preference.get('mode') != 'primary'
 
+    @property
+    def checksum(self):
+        """The provided checksum, if set, else None."""
+        return self._checksum
+
     slave_okay = slave_ok
     """Synonym for `.slave_ok`."""
 
@@ -1497,8 +1523,30 @@ class MockupDB(object):
             >>> client.close()
             >>> s.stop()
         """
+        return self._insert_responder("top", matcher, *args, **kwargs)
+
+    subscribe = autoresponds
+    """Synonym for `.autoresponds`."""
+
+    @_synchronized
+    def append_responder(self, matcher, *args, **kwargs):
+        """Add a responder of last resort.
+
+        Like `.autoresponds`, but instead of adding a responder to the top of
+        the stack, add it to the bottom. This responder will be called if no
+        others match.
+        """
+        return self._insert_responder("bottom", matcher, *args, **kwargs)
+
+    def _insert_responder(self, where, matcher, *args, **kwargs):
         responder = _AutoResponder(self, matcher, *args, **kwargs)
-        self._autoresponders.append(responder)
+        if where == "top":
+            self._autoresponders.append(responder)
+        elif where == "bottom":
+            self._autoresponders.insert(0, responder)
+        else:
+            raise RuntimeError("Invalid 'where': %r" % (where,))
+
         try:
             request = self._request_q.peek(block=False)
         except Empty:
@@ -1509,9 +1557,6 @@ class MockupDB(object):
 
         return responder
 
-    subscribe = autoresponds
-    """Synonym for `.autoresponds`."""
-
     @_synchronized
     def cancel_responder(self, responder):
         """Cancel a responder that was registered with `autoresponds`."""
@@ -1615,10 +1660,11 @@ class MockupDB(object):
                     server_thread.start()
             except socket.error as error:
                 if error.errno not in (
-                    errno.EAGAIN, errno.EBADF, errno.EWOULDBLOCK):
+                        errno.EAGAIN, errno.EBADF, errno.ENOTSOCK,
+                        errno.EWOULDBLOCK):
                     raise
             except select.error as error:
-                if error.args[0] == errno.EBADF:
+                if error.args[0] in (errno.EBADF, errno.ENOTSOCK):
                     # Closed.
                     break
                 else:
@@ -1643,12 +1689,13 @@ class MockupDB(object):
                 else:
                     self._request_q.put(request)
             except socket.error as error:
-                if error.errno in (errno.ECONNRESET, errno.EBADF):
+                if error.errno in (errno.ECONNRESET, errno.EBADF,
+                                   errno.ENOTSOCK):
                     # We hung up, or the client did.
                     break
                 raise
             except select.error as error:
-                if error.args[0] == errno.EBADF:
+                if error.args[0] in (errno.EBADF, errno.ENOTSOCK):
                     # Closed.
                     break
                 else:
@@ -2016,7 +2063,7 @@ def interactive_server(port=27017, verbose=True, all_ok=False, name='MockupDB',
                       auto_ismaster=True,
                       uds_path=uds_path)
     if all_ok:
-        server.autoresponds({})
+        server.append_responder({})
     server.autoresponds('whatsmyuri', you='localhost:12345')
     server.autoresponds({'getLog': 'startupWarnings'},
                         log=['hello from %s!' % name])
